from .classifier import *  # noqa:F401,F403
from .navigation import *  # noqa:F401,F403
from .project import *  # noqa:F401,F403
from .search import *  # noqa:F401,F403
from .sphinx_page import *  # noqa:F401,F403
from .version import *  # noqa:F401,F403



from django import forms

from .validators import validate_machine_name


class MachineNameField(forms.SlugField):
    """
    This is a form field for our
    :py:class:`sphinx_hosting.fields.MachineNameField` that applies the
    appropriate validators.
    """
    default_validators = [validate_machine_name]


from typing import Any, Dict, List

from django.template import Context, Template
from wildewidgets import (
    Block,
    CardHeader,
    CardWidget,
    Column,
    FontIcon,
    HTMLWidget,
    LinkButton,
    Menu,
    MenuItem,
    Row,
    TwoColumnLayout
)
from ..models import SphinxPage


#------------------------------------------------------
# SphinxPage related widgets
#------------------------------------------------------

class SphinxPagePagination(Row):
    """
    This widget draws the "Previous Page", Parent Page and Next Page buttons that
    are found at the top of each
    :py:class:`sphinx_hosting.views.SphinxPageDetailView`.

    It is built out of a Tabler/Bootstrap ``row``, with each of the buttons in
    an equal sized ``col``.
    """

    name: str = 'sphinxpage-pagination'

    def __init__(self, page: SphinxPage, **kwargs):
        super().__init__(**kwargs)
        self.add_column(Column(name='left', alignment='start', viewport_widths={'md': '4'}))
        self.add_column(Column(name='center', alignment='center', viewport_widths={'md': '4'}))
        self.add_column(Column(name='right', alignment='end', viewport_widths={'md': '4'}))
        if hasattr(page, 'previous_page') and page.previous_page.first():
            self.add_to_left(
                LinkButton(
                    text=Block(FontIcon('box-arrow-in-left'), page.previous_page.first().title),
                    url=page.previous_page.first().get_absolute_url(),
                    name=f'{self.name}__previous',
                    css_class='bg-azure-lt'
                )
            )
        if page.parent:
            self.add_to_center(
                LinkButton(
                    text=Block(FontIcon('box-arrow-in-up'), page.parent.title),
                    url=page.parent.get_absolute_url(),
                    name=f'{self.name}__parent',
                    css_class='bg-azure-lt'
                )
            )
        if page.next_page:
            self.add_to_right(
                LinkButton(
                    text=Block(page.next_page.title, FontIcon('box-arrow-in-right')),
                    url=page.next_page.get_absolute_url(),
                    name=f'{self.name}__next',
                    css_class='bg-azure-lt'
                )
            )


class SphinxPageBodyWidget(CardWidget):
    """
    This widget holds the body of the page.  The body as stored in the model is
    actually a Django template, so we retrieve the body, run it through the
    Django template engine, and display the results.

    Args:
        page: the ``SphinxPage`` we are rendering
    """

    css_class: str = 'sphinxpage-body'

    def __init__(self, page: SphinxPage, **kwargs):
        super().__init__(**kwargs)
        body = '{% load sphinx_hosting %}\n' + page.body
        self.widget = HTMLWidget(html=Template(body).render(Context()))


class SphinxPageTableOfContentsWidget(CardWidget):
    """
    This widget draws the in-page navigation -- the header hierarchy.

    Args:
        page: the ``SphinxPage`` we are rendering
    """

    css_class: str = 'sphinxpage-toc'

    def __init__(self, page: SphinxPage, **kwargs):
        super().__init__(**kwargs)
        self.widget = HTMLWidget(html=page.local_toc)
        self.set_header(
            CardHeader(
                header_level=3,
                header_text="Table of Contents",
                css_class=''
            )
        )


class SphinxPageGlobalTableOfContentsMenu(Menu):
    """
    This is the version-specific navigation menu that gets inserted into the
    page sidebar when viewing the documentation for a
    :py:class:`sphinx_hosting.models.Version`.  It will appear on all pages for
    that version.
    """

    css_class: str = 'mt-4'
    title_css_classes: str = 'mt-3'

    @classmethod
    def parse_obj(cls, data: Dict[str, Any]) -> "SphinxPageGlobalTableOfContentsMenu":
        """
        Given a dict like this::

            {
                items: [
                    {'text': 'foo'},
                    {'text': 'bar', 'url': '/foo', 'icon': 'blah'}
                    {'text': 'bar', 'url': '/foo', 'icon': 'blah', items: [{'text': 'blah' ...} ...]}
                    ...
                ]
            }

        Return a fully configured :py:class:`SphinxPageGlobalTableOfContentsMenu` suitable
        for insertion into a :py:class:`wildewidgets.Navbar`.

        Returns:
            A configured  ``SphinxPageGlobalTableOfContentsMenu``.
        """
        menu_items = cls._load_menuitems(data['items'])
        return cls(*menu_items)

    @classmethod
    def _load_menuitems(cls, items: List[Dict[str, Any]]) -> List[MenuItem]:
        """
        Given a list like this::

            [
               {'text': 'foo'},
               {'text': 'bar', 'url': '/foo', 'icon': 'blah'}
               {'text': 'bar', 'url': '/foo', 'icon': 'blah', items: [{'text': 'blah' ...} ...]}
                    ...
           ]

        Return a list of :py:class:`wildewidgets.MenuItem`
        objects loaded from that data.

        Returns:
            A list of :py:class:`wildewidgets.MenuItem` objects.
        """
        menu_items: List[MenuItem] = []
        for item in items:
            if 'items' in item:
                sub_items = cls._load_menuitems(item['items'])
                menu_items.append(MenuItem(
                    text=item['text'],
                    url=item.get('url', None),
                    icon=item.get('icon', None),
                    items=sub_items
                ))
            else:
                menu_items.append(MenuItem(**item))
        return menu_items


class SphinxPageTitle(Block):
    """
    The title block for a :py:class:`sphinx_hosting.models.SphinxPage` page.

    Args:
        page: the ``SphinxPage`` to render
    """

    block: str = 'sphinxpage-title'
    css_class: str = 'mb-5'

    def __init__(self, page: SphinxPage, **kwargs):
        super().__init__(**kwargs)
        self.project_title = page.version.project.title
        self.version_number = page.version.version
        self.title = page.title
        self.add_block(
            Block(
                f'{self.project_title}-{self.version_number}',
                name='project-title',
                css_class='text-muted fs-6 text-uppercase')
        )
        self.add_block(
            Block(
                self.title,
                tag='h1',
            )
        )


class SphinxPageLayout(Block):
    """
    The page layout for a single :py:class:`sphinx_hosting.models.SphinxPage`.
    It consists of a two column layout with the page's table of contents in the
    left column, and the content of the page in the right column.

    Args:
        page: the ``SphinxPage`` to render
    """

    left_column_width: int = 8

    def __init__(self, page: SphinxPage, **kwargs):
        super().__init__(**kwargs)
        self.add_block(SphinxPagePagination(page, css_class='mb-4'))
        self.add_block(SphinxPageTitle(page))
        layout = TwoColumnLayout(left_column_width=self.left_column_width)
        if page.local_toc:
            layout.add_to_right(SphinxPageTableOfContentsWidget(page))
        layout.add_to_left(SphinxPageBodyWidget(page))
        self.add_block(layout)
        self.add_block(SphinxPagePagination(page, css_class='mt-5'))



from django.db import models

from sphinx_hosting import form_fields

from .validators import validate_machine_name


class MachineNameField(models.SlugField):
    """
    This is just a :py:class:`django.forms.SlugField` that also allows "."
    characters.  "." is not uncommon in some project names, especially if
    the project is named after the website domain it hosts.
    """
    default_validators = [validate_machine_name]

    def formfield(self, **kwargs):
        return super().formfield(
            **{
                "form_class": form_fields.MachineNameField,
                "allow_unicode": self.allow_unicode,
                **kwargs,
            }
        )



from typing import Any, Dict

from django.conf import settings
from django.templatetags.static import static


app_settings: Dict[str, Any] = getattr(settings, 'SPHINX_HOSTING_SETTINGS', {})

LOGO_IMAGE: str = static(
    app_settings.get(
        'LOGO_IMAGE',
        'sphinx_hosting/images/logo.jpg',
    )
)
LOGO_URL: str = app_settings.get('LOGO_URL', '/')
LOGO_WIDTH: str = app_settings.get('LOGO_WIDTH', '100%')
SITE_NAME: str = app_settings.get('SITE_NAME', 'Sphinx Hosting')
MAX_GLOBAL_TOC_TREE_DEPTH: int = app_settings.get('MAX_GLOBAL_TOC_TREE_DEPTH', 2)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class SphinxHostingApiAppConfig(AppConfig):
    name: str = "sphinx_hosting.api"
    label: str = "sphinxhostingapi"
    verbose_name: str = _("Sphinx Hosting API")  # type: ignore
    default_auto_field: str = "django.db.models.AutoField"



from typing import Dict, List, Optional, Type, cast

from django.db.models import Model
from django.utils import timezone
from django.utils.html import strip_tags
from django.urls import reverse
from haystack.models import SearchResult
from haystack.query import SearchQuerySet
import humanize
from wildewidgets import (
    Block,
    Column,
    CrispyFormWidget,
    FontIcon,
    HorizontalLayoutBlock,
    Link,
    LinkButton,
    PagedModelWidget,
    Row,
    TagBlock
)

from ..forms import GlobalSearchForm
from ..models import Classifier, Project, SphinxPage


class GlobalSearchFormWidget(CrispyFormWidget):
    """
    This widget encapsulates the :py:class:`sphinx_hosting.forms.GlobalSearchForm`.
    """
    name: str = 'global-search'
    css_class: str = 'mb-3'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.form is None:
            self.form = GlobalSearchForm()


class SearchResultBlock(Block):
    """
    This is the block we use for rendering a particular search result on the search
    results page.

    Keyword Args:
        object: the :py:class:`haystack.models.SearchResult` object to render
    """

    block: str = 'search-result'
    max_text_length: int = 200

    class Header(Block):
        name: str = 'search-result__header'

        def __init__(self, result: SearchResult, **kwargs):
            super().__init__(**kwargs)
            self.add_class('mb-3')
            now = timezone.now()
            ago = humanize.naturaldelta(
                now - result.object.version.modified,
                minimum_unit='seconds'
            )
            self.add_block(
                HorizontalLayoutBlock(
                    Block(
                        str(result.object.version),
                        name='search-result__header__subtitle',
                        css_class='fs-6 text-muted font-bold'
                    ),
                    Block(
                        f'{ago} ago',
                        name='search-result__header__ago',
                        css_class='text-muted fs-6 text-uppercase'
                    ),
                    justify='between',
                    align='baseline'
                )
            )
            self.add_block(Block(result.object.title, tag='h3'))

    def __init__(self, object: SearchResult = None, **kwargs):  # pylint: disable=redefined-builtin
        result = cast(SearchResult, object)
        super().__init__(**kwargs)
        self.add_class('shadow')
        self.add_class('border')
        self.add_class('p-4')
        self.add_class('mb-4')
        self.add_block(SearchResultBlock.Header(result))
        page = cast(SphinxPage, result.object)
        text = strip_tags(page.body)
        text = text[:self.max_text_length].rsplit(' ', 1)[0] + '...'
        self.add_block(
            Block(text, name='search-result_snippet', css_class='fs-8 text-muted mb-3')
        )
        self.add_block(
            HorizontalLayoutBlock(
                LinkButton(text='Read', url=page.get_absolute_url()),
                Block(f'Rank: {result.score}', css_class='fs-6 text-muted'),
                justify='between',
                align='baseline'
            )
        )


class SearchResultsHeader(HorizontalLayoutBlock):
    """
    The header for the search results listing (not the page header -- that is
    :py:class:`SearchResultsPageHeader`).

    Args:
        results: the Haystack search queryset containing our search results
    """

    name: str = 'search-results__title'
    justify: str = 'between'

    def __init__(self, results: SearchQuerySet, **kwargs):
        super().__init__(**kwargs)
        self.add_block(
            Block(
                str(f"Search Results: {results.count()}"),
                css_class='fs-6 font-bold mb-3'
            )
        )


class PagedSearchResultsBlock(PagedModelWidget):
    """
    This is a paged listing of :py:class:`SearchResultBlock` entries describing
    our search results.
    """

    model: Type[Model] = SphinxPage
    page_kwarg: str = 'p'
    paginate_by: int = 10
    model_widget: Block = SearchResultBlock

    def __init__(self, results: SearchQuerySet, query: Optional[str], **kwargs):
        if query is not None:
            kwargs['extra_url'] = {'q': query}
        super().__init__(queryset=results, **kwargs)


class FacetBlock(Block):
    """
    This is a base class for blocks that appear to the right of the search
    results listing on the search results page that allows you to refine your
    results by a particular facet that is present in the result set.

    Subclass this to create facet filtering blocks for specific facets.  Any
    facet you want to filter by must be defined on your search index by adding
    ``faceted=True`` to the field definition.

    Args:
        results: the Haystack search queryset containing our search results
        query: the text entered into the search form that got us to this results
            page
    """

    #: The model class which our facet is related to
    model: Type[Model]
    #: The title for our block
    title: str
    #: The name of the facet
    facet: str
    #: The field on :py:attr:`model` that we will filter by
    model_field: str

    def __init__(self, results: SearchQuerySet, query: Optional[str], **kwargs):
        self.query = query
        super().__init__(**kwargs)
        self.add_class('border')
        self.add_class('bg-white')
        facet_qs = results.facet(self.facet)
        stats = facet_qs.facet_counts()
        self.add_block(Block(self.title, tag='h3', css_class='p-3'))
        body = Block(name='list-group', css_class='list-group-flush')
        for identifier, count in stats['fields'][self.facet]:
            kwargs = {self.model_field: identifier}
            instance = self.model.objects.get(**kwargs)
            body.add_block(
                Block(
                    HorizontalLayoutBlock(
                        Link(str(instance), url=instance.get_absolute_url(), css_class='fs-5'),  # type: ignore
                        HorizontalLayoutBlock(
                            TagBlock(count, color='cyan', css_class='me-2'),
                            LinkButton(
                                text='Filter',
                                url=reverse('sphinx_hosting:search') + f"?q={query}&{self.facet}={identifier}",
                                color='outline-secondary',
                                size='sm'
                            ),
                        )
                    ),
                    tag='li',
                    name='list-group-item'
                )
            )
        self.add_block(body)


class SearchResultsClassifiersFacet(FacetBlock):
    """
    A :py:class:`FacetBlock` that allows the user to filter search results by classifier.
    """

    model: Type[Model] = Classifier
    title: str = 'Classifiers'
    facet: str = 'classifiers'
    model_field: str = 'name'


class SearchResultsProjectFacet(FacetBlock):
    """
    A :py:class:`FacetBlock` that allows the user to filter search results by project.
    """

    model: Type[Model] = Project
    title: str = 'Projects'
    facet: str = 'project_id'
    model_field: str = 'pk'


class SearchResultsPageHeader(Block):
    """
    The header for the entire search results page.  This shows the search string
    that got us here, and any active facet filters, allowing the user to remove
    any active filter by clicking the "X" next to the filter name.

    Args:
        query: the text entered into the search form that got us to this results
            page

    Keyword Args:
        facets: the active facet filters.  This will be a dict where the key is the
            facet name, and the value is a list of facet values to filter by.
    """

    block: str = 'search-results__header'
    css_class: str = 'mb-5'

    def __init__(
        self,
        query: Optional[str],
        facets: Optional[Dict[str, List[str]]] = None,
        **kwargs
    ):
        if facets is None:
            facets = {}
        super().__init__(**kwargs)
        self.add_class('mb-4')
        self.add_block(
            Block(
                "Search Results",
                tag='h1',
                name='search-results__header__title',
            )
        )
        self.add_block(
            Block(
                f'Query: "{query}"',
                name='search-results__header__subtitle',
                css_class='text-muted fs-6 text-uppercase'
            )
        )
        if facets:
            buttons = HorizontalLayoutBlock(
                Block('Filters:', tag='h3', css_class='me-3'),
                justify='start',
                align='baseline',
                css_class='mt-3'
            )
            for facet, identifiers in facets.items():
                for identifier in identifiers:
                    if facet == 'project_id':
                        project = Project.objects.get(pk=identifier)
                        label = Block(
                            FontIcon(icon='file-excel-fill'),
                            f'Project: {project.title}'
                        )
                    elif facet == 'classifiers':
                        classifier = Classifier.objects.get(name=identifier)
                        label = Block(
                            FontIcon(icon='file-excel-fill'),
                            f'Classifier: {classifier.name}'
                        )
                    buttons.add_block(
                        LinkButton(
                            text=label,
                            url=reverse('sphinx_hosting:search') + f"?q={query}",
                            color='outline-azure',
                            css_class='me-3'
                        )
                    )
            self.add_block(buttons)


class PagedSearchLayout(Block):
    """
    This is the page layout for the entire search results page.

    Args:
        query: the text entered into the search form that got us to this results
            page


    Keyword Args:
        query: the text entered into the search form that got us to this results
            page
        facets: the active facet filters.  This will be a dict where the key is
            the facet name, and the value is a list of facet values to filter by.
    """

    name: str = 'search-layout'
    modifier: str = 'paged'

    def __init__(
        self,
        results: SearchQuerySet,
        query: Optional[str] = None,
        facets: Optional[Dict[str, List[str]]] = None,
        **kwargs
    ):
        self.query = query
        if facets is None:
            facets = {}
        super().__init__(**kwargs)
        self.add_block(SearchResultsPageHeader(query, facets=facets))
        self.add_block(SearchResultsHeader(results))
        row = Row()
        row.add_column(
            Column(
                PagedSearchResultsBlock(results, query),
                name='middle',
                base_width=8
            )
        )
        row.add_column(
            Column(
                SearchResultsProjectFacet(results, query, css_class='mb-4'),
                SearchResultsClassifiersFacet(results, query),
                name='right',
                base_width=4
            )
        )
        self.add_block(row)



from typing import List, Union

from django.urls import include, path, re_path, URLPattern, URLResolver

from drf_spectacular.views import (
    SpectacularAPIView,
    SpectacularRedocView,
    SpectacularSwaggerView
)
from rest_framework import routers

from .views import (
    ClassifierViewSet,
    ProjectViewSet,
    VersionViewSet,
    SphinxPageViewSet,
    SphinxImageViewSet,
    VersionUploadView
)

app_name: str = "sphinx_hosting_api"
router: routers.BaseRouter = routers.DefaultRouter()

# Users
router.register(r'classifiers', ClassifierViewSet)
router.register(r'projects', ProjectViewSet)
router.register(r'versions', VersionViewSet)
router.register(r'pages', SphinxPageViewSet)
router.register(r'images', SphinxImageViewSet)

urlpatterns: List[Union[URLPattern, URLResolver]] = [
    path('', include(router.urls)),
    path('version/import/', VersionUploadView.as_view(), name='version-import'),
    path('schema/', SpectacularAPIView.as_view(), name='schema'),
    path(
        'schema/swagger-ui/',
        SpectacularSwaggerView.as_view(url_name='sphinx_hosting_api:schema'),
        name='swagger-ui'
    ),
    path(
        'schema/redoc/',
        SpectacularRedocView.as_view(url_name='sphinx_hosting_api:schema'),
        name='redoc'
    ),
]



from typing import Dict, List, Optional, Type

from django.db.models import Model, QuerySet
from django.db.models.functions import Length
from wildewidgets import (
    Block,
    BasicModelTable,
    CardWidget,
    CrispyFormWidget,
    Datagrid,
    WidgetListLayoutHeader
)

from ..models import (
    SphinxImage,
    SphinxPage,
    Version
)


class VersionInfoWidget(CardWidget):
    """
    This widget gives a :py:class:`wildewidget.Datagrid` type overview of
    information about this version:

    * A link to the project that owns this
      :py:class:`sphinx_hosting.models.Version`
    * Create and last modified timestamps
    * What version of Sphinx was used to generate the pages

    Args:
        version: the ``Version`` object we're describing
    """

    title: str = "Version Info"
    icon: str = "info-square"

    def __init__(self, version: Version, **kwargs):
        super().__init__(**kwargs)
        grid = Datagrid()
        grid.add_item(
            title='Project',
            content=version.project.title,
            url=version.project.get_absolute_url()
        )
        grid.add_item(title='Version Created', content=version.created.strftime('%Y-%m-%d %H:%M %Z'))
        grid.add_item(title='Version Last Modified', content=version.modified.strftime('%Y-%m-%d %H:%M %Z'))
        grid.add_item(title='Sphinx Version', content=version.sphinx_version)
        self.set_widget(grid)


class VersionSphinxPageTableWidget(CardWidget):
    """
    This is a :py:class:`wildewidgets.CardWidget` that gives our
    :py:class:`VersionSphinxPageTable` dataTable a nice header with a total
    page count.
    """
    title: str = "Pages"
    icon: str = "bookmark-star"

    def __init__(self, version_id: int, **kwargs):
        self.version_id = version_id
        super().__init__(
            widget=VersionSphinxPageTable(version_id=version_id),
            **kwargs,
        )

    def get_title(self) -> WidgetListLayoutHeader:
        header = WidgetListLayoutHeader(
            header_text="Pages",
            badge_text=Version.objects.get(pk=self.version_id).pages.count(),
        )
        return header


class VersionUploadBlock(CardWidget):
    """
    This block holds the upload form for uploading documentation tarballs.  Once
    uploaded, the tarball will be run through
    :py:class:`sphinx_hosting.importers.SphinxPackageImporter` to actually
    import it into the database.
    """

    css_class: str = "my-3 border"

    def __init__(self, *blocks, form=None, **kwargs):
        super().__init__(
            widget=CrispyFormWidget(form=form, name='project__upload_docs'),
            **kwargs
        )
        self.set_header(Block("Import Docs", tag='h3'))


class VersionSphinxImageTableWidget(CardWidget):
    """
    This is a :py:class:`wildewidgets.CardWidget` that gives our
    :py:class:`VersionSphinxImageTable` dataTable a nice header with a total
    image count.
    """
    title: str = "Images"
    icon: str = "bookmark-star"

    def __init__(self, version_id: int, **kwargs):
        self.version_id = version_id
        super().__init__(
            widget=VersionSphinxImageTable(version_id=version_id),
            **kwargs,
        )

    def get_title(self) -> WidgetListLayoutHeader:
        header = WidgetListLayoutHeader(
            header_text="Images",
            badge_text=Version.objects.get(pk=self.version_id).images.count(),
        )
        return header


#------------------------------------------------------
# Datatables
#------------------------------------------------------

class VersionSphinxPageTable(BasicModelTable):
    """
    This widget displays a `dataTable <https://datatables.net>`_ of our
    :py:class:`sphinx_hosting.models.SphinxPage` instances for a particular
    :py:class:`sphinx_hosting.models.Version`.

    It's used as a the main widget in by :py:class:`VersionSphinxPageTableWidget`.
    """

    model: Type[Model] = SphinxPage

    page_length: int = 25  #: Show this many books per page
    striped: bool = True   #: Set to ``True`` to stripe our table rows
    actions: bool = True

    fields: List[str] = [  #: These are the fields on our model (or which are computed) that we will list as columns
        'title',
        'relative_path',
        'size',
    ]
    alignment: Dict[str, str] = {  #: declare how we horizontally align our columns
        'title': 'left',
        'relative_path': 'left',
        'size': 'right',
    }

    def __init__(self, *args,  **kwargs) -> None:
        """
        One of our ``kwargs`` must be ``version_id``, the ``pk`` of the
        :py:class:`sphinx_hosting.models.Version` for which we want to list
        :py:class:`sphinx_hosting.models.SphinxPage` objects.

        This will get added to the :py:attr:`extra_data` dict in the ``kwargs``
        key, from which we reference it.
        """
        #: The pk of the :py:class:`sphinx_hosting.models.Version` for which to list pages
        self.version_id: Optional[int] = kwargs.get('version_id', None)
        super().__init__(*args, **kwargs)
        if 'version_id' in self.extra_data['kwargs']:
            self.version_id = int(self.extra_data['kwargs']['version_id'])

    def get_initial_queryset(self) -> QuerySet[SphinxPage]:
        """
        Filter our :py:class:`sphinx_hosting.models.SphinxPage` objects by
        :py:attr:`version_id`.
        """
        qs = (
            super().get_initial_queryset()
            .filter(version_id=self.version_id)
            .annotate(size=Length('body'))
        )
        return qs.order_by('title')


class VersionSphinxImageTable(BasicModelTable):
    """
    This widget displays a `dataTable <https://datatables.net>`_ of our
    :py:class:`sphinx_hosting.models.SphinxImage` instances for a particular
    :py:class:`sphinx_hosting.models.Version`.

    It's used as a the main widget in by :py:class:`VersionSphinxImageTableWidget`.
    """

    model: Type[Model] = SphinxImage

    page_length: int = 25  #: Show this many books per page
    striped: bool = True   #: Set to ``True`` to stripe our table rows

    fields: List[str] = [  #: These are the fields on our model (or which are computed) that we will list as columns
        'orig_path',
        'file_path',
        'size',
    ]
    alignment: Dict[str, str] = {  #: declare how we horizontally align our columns
        'orig_path': 'left',
        'file_path': 'left',
        'size': 'right',
    }

    def __init__(self, *args, **kwargs) -> None:
        """
        One of our ``kwargs`` must be ``version_id``, the ``pk`` of the
        :py:class:`sphinx_hosting.models.Version` for which we want to list
        :py:class:`sphinx_hosting.models.SphinxPage` objects.

        This will get added to the :py:attr:`extra_data` dict in the ``kwargs``
        key, from which we reference it.
        """
        #: The pk of the :py:class:`sphinx_hosting.models.Version` for which to list pages
        self.version_id: Optional[int] = kwargs.get('version_id', None)
        super().__init__(*args, **kwargs)
        if 'version_id' in self.extra_data['kwargs']:
            self.version_id = int(self.extra_data['kwargs']['version_id'])

    def get_initial_queryset(self) -> QuerySet[SphinxPage]:
        """
        Filter our :py:class:`sphinx_hosting.models.SphinxPage` objects by
        :py:attr:`version_id`.
        """
        qs = (
            super().get_initial_queryset()
            .filter(version_id=self.version_id)
        )
        return qs.order_by('orig_path')

    def render_size_column(self, row: Version, column: str) -> str:
        """
        Render our ``size`` column.  This is the size in bytes of the
        :py:attr:`sphinx_hosting.models.SphinxImage.file` field.

        Args:
            row: the ``Version`` we are rendering
            colunn: the name of the column to render

        Returns:
            The size in bytes of the uploaded file.
        """
        return str(row.file.size)

    def render_file_path_column(self, row: Version, column: str) -> str:
        """
        Render our ``file_path`` column.  This is the path to the file in
        ``MEDIA_ROOT``.

        Args:
            row: the ``Version`` we are rendering
            colunn: the name of the column to render

        Returns:
            The size in bytes of the uploaded file.
        """
        return str(row.file.name)



from setuptools import setup, find_packages

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name="django-sphinx-hosting",
    version="1.1.3",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "django-braces >= 1.15.0",
        "django-crispy-forms>=1.14.0",
        "django-extensions >= 3.2.1",
        "django-filter >= 22.1",
        "django-haystack >= 3.2.1",
        "django-rest-framework-helpers >= 8.5.0",
        "django-theme-academy >= 0.3.0",
        "django-wildewidgets >= 0.16.7",
        "djangorestframework >= 3.14.0",
        "drf-spectacular >= 0.25.1",
        "crispy-bootstrap5",
        "humanize",
        "lxml >= 4.9.1",
        "cssselect >= 1.2.0",
        "rich"
    ],
    author="Caltech IMSS ADS",
    author_email="imss-ads-staff@caltech.edu",
    url="https://github.com/caltechads/django-sphinx-hosting",
    description="Reusable Django app for hosting Sphinx documentation privately.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=['documentation', 'sphinx', 'django'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.11',
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Framework :: Django :: 4.1",
        "Topic :: Documentation :: Sphinx",
        "Topic :: Software Development :: Documentation",
    ],
)



from typing import Type, Tuple, Dict, Any

from django.db.models import Model
from rest_framework import serializers
from rest_framework.relations import RelatedField

from sphinx_hosting.models import (
    Classifier,
    Project,
    SphinxImage,
    SphinxPage,
    Version,
)


class ClassifierSerializer(serializers.HyperlinkedModelSerializer):

    class Meta:
        model: Type[Model] = Classifier
        fields: Tuple[str, ...] = (
            'url',
            'id',
            'name'
        )
        extra_kwargs = {
            'url': {'view_name': 'sphinx_hosting_api:classifier-detail'},
        }


class ProjectSerializer(serializers.HyperlinkedModelSerializer):

    versions: RelatedField = serializers.HyperlinkedRelatedField(
        many=True,
        read_only=True,
        view_name='sphinx_hosting_api:version-detail'
    )
    classifiers: serializers.Serializer = ClassifierSerializer(many=True)

    class Meta:
        model: Type[Model] = Project
        read_only_fields: Tuple[str, ...] = ('machine_name', )
        fields: Tuple[str, ...] = (
            'url',
            'id',
            'title',
            'machine_name',
            'description',
            'classifiers',
            'versions',
        )
        extra_kwargs = {
            'url': {'view_name': 'sphinx_hosting_api:project-detail'},
        }


class VersionSerializer(serializers.HyperlinkedModelSerializer):

    project: RelatedField = serializers.HyperlinkedRelatedField(
        read_only=True,
        view_name='sphinx_hosting_api:project-detail'
    )
    head: RelatedField = serializers.HyperlinkedRelatedField(
        read_only=True,
        view_name='sphinx_hosting_api:sphinxpage-detail'
    )
    pages: RelatedField = serializers.HyperlinkedRelatedField(
        many=True,
        read_only=True,
        view_name='sphinx_hosting_api:sphinxpage-detail'
    )
    images: RelatedField = serializers.HyperlinkedRelatedField(
        many=True,
        read_only=True,
        view_name='sphinx_hosting_api:sphinximage-detail'
    )

    class Meta:
        model: Type[Model] = Version
        fields: Tuple[str, ...] = (
            'url',
            'id',
            'project',
            'version',
            'sphinx_version',
            'archived',
            'head',
            'pages',
            'images',
        )
        extra_kwargs = {
            'url': {'view_name': 'sphinx_hosting_api:version-detail'},
        }


class VersionUploadSerializer(serializers.Serializer):
    """
    The actual work of importing the file is done in
    :py:class:`sphinx_hosting.api.views.VersionUploadView`.  We're defining our
    :py:meth:`create` and :py:meth:`create` here as NOOP functions to make the
    linters happy because they're abstract in :py:class:`serializers.Serializer`.
    """

    file = serializers.FileField()

    def create(self, validated_data: Dict[str, Any]) -> None:
        pass

    def update(self, instance, validated_data):
        pass

    class Meta:
        fields = ('file', )


class SphinxPageSerializer(serializers.HyperlinkedModelSerializer):

    version: RelatedField = serializers.HyperlinkedRelatedField(
        read_only=True,
        view_name='sphinx_hosting_api:version-detail'
    )
    parent: RelatedField = serializers.HyperlinkedRelatedField(   # type: ignore
        read_only=True,
        view_name='sphinx_hosting_api:sphinxpage-detail'
    )
    previous_page: RelatedField = serializers.HyperlinkedRelatedField(
        many=True,
        read_only=True,
        view_name='sphinx_hosting_api:sphinxpage-detail'
    )
    next_page: RelatedField = serializers.HyperlinkedRelatedField(
        read_only=True,
        view_name='sphinx_hosting_api:sphinxpage-detail'
    )

    class Meta:
        model: Type[Model] = SphinxPage
        fields: Tuple[str, ...] = (
            'url',
            'id',
            'version',
            'title',
            'relative_path',
            'content',
            'orig_body',
            'body',
            'orig_local_toc',
            'local_toc',
            'orig_global_toc',
            'searchable',
            'parent',
            'next_page',
            'previous_page',
        )
        extra_kwargs = {
            'url': {'view_name': 'sphinx_hosting_api:sphinxpage-detail'},
        }


class SphinxImageSerializer(serializers.HyperlinkedModelSerializer):
    version: RelatedField = serializers.HyperlinkedRelatedField(
        many=True,
        read_only=True,
        view_name='sphinx_hosting_api:version-detail'
    )

    class Meta:
        model: Type[Model] = SphinxImage
        fields: Tuple[str, ...] = (
            'url',
            'id',
            'version',
            'orig_path',
        )
        extra_kwargs = {
            'url': {'view_name': 'sphinx_hosting_api:sphinximage-detail'},
        }



# Generated by Django 4.1.7 on 2023-03-27 18:19

import itertools
from django.db import migrations
from django.contrib.auth.management import create_permissions


def apply_migration(apps, schema_editor):
    # Force the creation of the permission objects we'll need
    for app_config in apps.get_app_configs():
        app_config.models_module = True
        create_permissions(app_config, verbosity=0)
        app_config.models_module = None
    db_alias = schema_editor.connection.alias
    Group = apps.get_model("auth", "Group")
    Permission = apps.get_model("auth", "Permission")
    # Create our groups
    Group.objects.using(db_alias).bulk_create(
        [
            Group(name="Administrators"),
            Group(name="Classifier Managers"),
            Group(name="PermissionGroup Managers"),
            Group(name="Project Managers"),
            Group(name="Version Managers"),
            Group(name="Editors"),
        ]
    )
    # Now assign permissions
    group = Group.objects.using(db_alias).get(name="Classifier Managers")
    classifier_permissions = Permission.objects.filter(codename__endswith='classifier')
    group.permissions.set(classifier_permissions)

    group = Group.objects.using(db_alias).get(name="Project Managers")
    project_permissions = Permission.objects.filter(codename__endswith='project')
    group.permissions.set(project_permissions)

    group = Group.objects.using(db_alias).get(name="PermissionGroup Managers")
    permissiongroup_permissions = Permission.objects.filter(codename__endswith='permissiongroup')
    group.permissions.set(permissiongroup_permissions)

    group = Group.objects.using(db_alias).get(name="Version Managers")
    version_permissions = Permission.objects.filter(codename__endswith='version')
    group.permissions.set(version_permissions)

    group = Group.objects.using(db_alias).get(name="Administrators")
    group.permissions.set(
        itertools.chain(
            classifier_permissions,
            permissiongroup_permissions,
            project_permissions,
            version_permissions
        )
    )
    group = Group.objects.using(db_alias).get(name="Editors")
    group.permissions.set(
        itertools.chain(
            project_permissions,
            version_permissions
        )
    )


def revert_migration(apps, schema_editor):
    Group = apps.get_model("auth", "Group")
    Group.objects.filter(
        name__in=[
            "Classifier Managers",
            "PermissionGroup Managers",
            "Project Managers",
            "Version Managers",
            "Editors",
            "Administrators",
        ]
    ).delete()


class Migration(migrations.Migration):

    dependencies = [
        ('sphinxhostingcore', '0009_SphinxPage_searchable'),
    ]

    operations = [migrations.RunPython(apply_migration, revert_migration)]



# Generated by Django 4.1.7 on 2023-05-03 17:01

from django.db import migrations, models
import sphinx_hosting.fields


class Migration(migrations.Migration):

    dependencies = [
        ('sphinxhostingcore', '0010_add_groups'),
    ]

    operations = [
        migrations.AlterField(
            model_name='project',
            name='machine_name',
            field=sphinx_hosting.fields.MachineNameField(help_text='Must be unique.  Set this to the slugified value of "project" in Sphinx\'s. conf.py', unique=True, verbose_name='Machine Name'),
        ),
        migrations.AlterField(
            model_name='sphinxpage',
            name='orig_body',
            field=models.TextField(blank=True, help_text='The original body for the page, extracted from the page JSON. Some pages have no body. We save this here in case we need to reprocess the body at some later date.', verbose_name='Body (Original)'),
        ),
    ]



from typing import Dict, List, Optional, Type

from django.contrib.auth.models import AbstractUser
from django.db.models import Model, QuerySet
from wildewidgets import (
    ActionButtonModelTable,
    BasicModelTable,
    Block,
    CrispyFormModalWidget,
    CrispyFormWidget,
    CardWidget,
    Datagrid,
    ListModelWidget,
    RowActionButton,
    RowEditButton,
    RowModelUrlButton,
    ToggleableManyToManyFieldBlock,
    TwoColumnLayout,
    Widget,
    WidgetListLayoutHeader,
)

from ..forms import ProjectCreateForm
from ..models import Classifier, Project, Version

from .classifier import ClassifierFilterBlock


#------------------------------------------------------
# Modals
#------------------------------------------------------

class ProjectCreateModalWidget(CrispyFormModalWidget):
    """
    This is a modal dialog that holds the
    :py:class:`sphinx_hosting.forms.ProjectCreateForm`.
    """

    modal_id: str = "project__create"
    modal_title: str = "Add Project"

    def __init__(self, *args, **kwargs):
        form = ProjectCreateForm()
        super().__init__(form=form, *args, **kwargs)


#------------------------------------------------------
# Project related widgets
#------------------------------------------------------

class ProjectInfoWidget(CardWidget):
    """
    This is a :py:class:`wildewidgets.CardWidget` containing a Tabler datagrid
    that gives information about a :py:class:`sphinx_hosting.models.Project`.
    """

    title: str = "Project Info"
    icon: str = "info-square"

    def __init__(self, project: Project, **kwargs):
        super().__init__(**kwargs)
        grid = Datagrid()
        grid.add_item(title='Machine Name', content=project.machine_name)
        grid.add_item(title='Created', content=project.created.strftime('%Y-%m-%d %H:%M %Z'))
        grid.add_item(title='Last Modified', content=project.modified.strftime('%Y-%m-%d %H:%M %Z'))
        self.set_widget(grid)


class ProjectTableWidget(Block):
    """
    This is a :py:class:`wildewidgets.CardWidget` that gives our
    :py:class:`ProjectTable` dataTable a nice header with a total book count and
    an "Add Project" button that opens a modal dialog.
    """

    def __init__(self, user: AbstractUser, **kwargs):
        super().__init__(**kwargs)
        self.add_block(self.get_title(user))
        layout = TwoColumnLayout(left_column_width=9)
        table = ProjectTable()
        layout.add_to_left(CardWidget(widget=table))
        layout.add_to_right(
            ClassifierFilterBlock(
                table_name=table.table_name,
                column_number=table.get_column_number('classifiers')
            )
        )
        self.add_block(layout)

    def get_title(self, user: AbstractUser) -> WidgetListLayoutHeader:
        header = WidgetListLayoutHeader(
            header_text="Projects",
            badge_text=Project.objects.count(),
        )
        if user.has_perm('sphinxhostingcore.add_project'):
            header.add_modal_button(
                text="New Project",
                color="primary",
                target=f'#{ProjectCreateModalWidget.modal_id}'
            )
        return header


class ProjectVersionsTableWidget(CardWidget):
    """
    This is a :py:class:`wildewidgets.CardWidget` that gives our
    :py:class:`ProjectVersionTable` dataTable a nice header with a total version
    count.
    """
    title: str = "Versions"
    icon: str = "bookmark-star"

    def __init__(self, project_id: int, **kwargs):
        self.project_id = project_id
        super().__init__(
            widget=ProjectVersionTable(project_id=project_id),
            **kwargs,
        )

    def get_title(self) -> WidgetListLayoutHeader:
        header = WidgetListLayoutHeader(
            header_text="Versions",
            badge_text=Project.objects.get(pk=self.project_id).versions.count(),
        )
        return header


class ProjectClassifierSelectorWidget(ToggleableManyToManyFieldBlock):

    model = Project
    field_name = 'classifiers'
    title = 'Classifiers'
    icon = 'collection'


class ProjectClassifierListWidget(ListModelWidget):

    paginate_by: int = 100
    item_label: str = 'Classifier'
    title = 'Classifiers'
    icon = 'collection'

    def get_object_text(self, instance: Classifier) -> str:
        return instance.name

    def get_model_subblock(self, instance: Classifier) -> Block:
        return Block(self.get_object_text(instance), tag='label')


class ProjectDetailWidget(
    CrispyFormWidget,
    Widget
):
    """
    This widget renders an update form for a :py:class:`sphinx_hosting.models.Project`.

    Use directly it like so::

        >>> project = Project.objects.get(pk=1)
        >>> form = ProjectUpdateForm(instance=project)
        >>> widget = ProjectDetailWidget(form=form)

    Or you can simply add the form to your view context and
    :py:class:`ProjectDetailWidget` will pick it up automatically.
    """
    title: str = "General Settings"
    name: str = 'project-detail__section'
    modifier: str = 'general'
    icon: str = "card-checklist"
    css_class: str = CrispyFormWidget.css_class + " p-4"


#------------------------------------------------------
# Datatables
#------------------------------------------------------

class LatestVersionButton(RowModelUrlButton):

    attribute: str = 'get_latest_version_url'
    text: str = 'Read Docs'
    color: str = 'orange'

    def is_visible(self, row: Project, user: AbstractUser) -> bool:
        if row.latest_version is None:
            return False
        return super().is_visible(row, user)


class ProjectTable(ActionButtonModelTable):
    """
    This widget displays a `dataTable <https://datatables.net>`_ of our
    :py:class:`sphinx_hosting.models.Project` instances.

    It's used as a the main widget in by :py:class:`ProjectTableWidget`.
    """

    model: Type[Model] = Project

    #: Show this many rows per page
    page_length: int = 25
    #: Set to ``True`` to stripe our table rows
    striped: bool = True
    default_action_button_label = 'Edit'
    default_action_button_color_class = 'outline-secondary'
    #: A list of fields that we will list as columns.  These are either fields
    #: on our :py:attr:`model`, or defined as ``render_FIELD_NAME_column`` methods
    #: on this class
    fields: List[str] = [
        'title',
        'machine_name',
        'classifiers',
        'latest_version',
        'latest_version_date',
    ]
    #: A list of names of columns to hide by default.
    hidden: List[str] = [
        'classifiers',
        'machine_name'
    ]
    #: A list of names of columns that will will not be searched when doing a
    #: **global** search
    unsearchable: List[str] = [
        'lastest_version',
        'latest_version_date',
    ]
    #: A dict of column name to column label.  We use it to override the
    #: default labels for the named columns
    verbose_names: Dict[str, str] = {
        'title': 'Project Name',
        'machine_name': 'Machine Name',
        'latest_version': 'Latest Version',
        'latest_version_date': 'Import Date',
    }
    #: A dict of column names to alignment ("left", "right", "center")
    alignment: Dict[str, str] = {
        'title': 'left',
        'classifiers': 'left',
        'machine_name': 'left',
        'latest_version': 'left',
        'latest_version_date': 'left'
    }

    actions: List[RowActionButton] = [
        LatestVersionButton(),
        RowModelUrlButton(
            attribute='get_absolute_url',
            text='View',
            color='outline-secondary'
        ),
        RowEditButton(
            permission='sphinxhostingcore.change_project',
            color='azure'
        ),
    ]

    def render_latest_version_column(self, row: Project, column: str) -> str:
        """
        Render our ``latest_version`` column.  This is the version string of the
        :py:class:`sphinx_hosting.models.Version` that has the most recent
        :py:attr:`sphinx_hosting.models.Version.modified` timestamp.

        If there are not yet any :py:class:`sphinx_hosting.models.Version` instances for
        this project, return empty string.

        Args:
            row: the ``Project`` we are rendering
            colunn: the name of the column to render

        Returns:
            The version string of the most recently published version, or empty
            string.
        """
        version = row.versions.order_by('-modified').first()
        if version:
            return version.version
        return ''

    def render_latest_version_date_column(self, row: Project, column: str) -> str:
        """
        Render our ``latest_version_date`` column.  This is the last modified
        date of the :py:class:`sphinx_hosting.models.Version` that has the most
        recent :py:attr:`sphinx_hosting.models.Version.modified` timestamp.

        If there are not yet any :py:class:`sphinx_hosting.models.Version`
        instances for this project, return empty string.

        Args:
            row: the ``Project`` we are rendering
            colunn: the name of the column to render

        Returns:
            The of the most recently published version, or empty
            string.
        """
        version = row.versions.order_by('-modified').first()
        if version:
            return self.render_datetime_type_column(version.modified)
        return ''

    def render_classifiers_column(self, row: Project, column: str) -> str:
        """
        Render our ``classifiers`` column.

        Args:
            row: the ``Project`` we are rendering
            colunn: the name of the column to render

        Returns:
            A ``<br>`` separated list of classifier names
        """
        return '<br>'.join(row.classifiers.values_list('name', flat=True))

    def filter_classifiers_column(
        self,
        qs: QuerySet,
        column: str,
        value: str
    ) -> QuerySet:
        """
        Filter our results by the ``value``, a comma separated list of
        :py:class:`sphinx_hosting.models.Classifier` names.

        Args:
            qs: the current :py:class:`QuerySet`
            colunn: the name of the column to filter on
            value: a comma-separated list of classifier names

        Returns:
            A :py:class:`QuerySet` filtered for rows that contain the selected
            classifiers.
        """
        classifier_ids = value.split(',')
        return qs.filter(classifiers__id__in=classifier_ids).distinct()


class ProjectVersionTable(BasicModelTable):
    """
    This widget displays a `dataTable <https://datatables.net>`_ of our
    :py:class:`sphinx_hosting.models.Version` instances for a particular
    :py:class:`sphinx_hosting.models.Project`.

    It's used as a the main widget in by :py:class:`ProjectVersionTableWidget`.
    """

    model: Type[Model] = Version

    #: Show this many rows per page
    page_length: int = 25
    #: Set to ``True`` to stripe our table rows
    striped: bool = True
    actions: bool = True

    #: A list of fields that we will list as columns.  These are either fields
    #: on our :py:attr:`model`, or defined as ``render_FIELD_NAME_column`` methods
    #: on this class
    fields: List[str] = [
        'version',
        'num_pages',
        'num_images',
        'created',
        'modified',
    ]
    #: A list of names of columns that will will not be searched when doing a
    #: **global** search
    unsearchable: List[str] = [
        'num_pages',
        'num_images',
    ]
    #: A dict of column name to column label.  We use it to override the
    #: default labels for the named columns
    verbose_names: Dict[str, str] = {
        'title': 'Version',
        'num_pages': '# Pages',
        'num_images': '# Images',
    }
    #: A dict of column names to alignment ("left", "right", "center")
    alignment: Dict[str, str] = {
        'version': 'left',
        'num_pages': 'right',
        'num_images': 'right',
        'created': 'left',
        'modified': 'left'
    }
    #: Order by version number
    order_columns: List[str] = ['version']
    #: Sort so that the highest version number is on top
    sort_ascending: bool = False

    def __init__(self, *args, **kwargs) -> None:
        """
        One of our ``kwargs`` must be ``project_id``, the ``pk`` of the
        :py:class:`sphinx_hosting.models.Project` for which we want to list
        :py:class:`sphinx_hosting.models.Version` objects.

        This will get added to the :py:attr:`extra_data` dict in the ``kwargs``
        key, from which we reference it.
        """
        #: The pk of the :py:class:`sphinx_hosting.models.Project` for which to list versions
        self.project_id: Optional[int] = kwargs.get('project_id', None)
        super().__init__(*args, **kwargs)
        if 'project_id' in self.extra_data['kwargs']:
            self.project_id = int(self.extra_data['kwargs']['project_id'])

    def get_initial_queryset(self) -> QuerySet[Version]:
        """
        Filter our :py:class:`sphinx_hosting.models.Version` objects by
        :py:attr:`project_id`.

        Returns:
            A filtered :py:class:`QuerySet` on :py:class:`sphinx_hosting.models.Version`
        """
        return super().get_initial_queryset().filter(project_id=self.project_id)

    def render_num_pages_column(self, row: Version, column: str) -> str:
        """
        Render our ``num_pages`` column.  This is the number of
        :py:class:`sphinx_hosting.models.SphinxPage` objects imported for this
        version.

        Args:
            row: the ``Version`` we are rendering
            colunn: the name of the column to render

        Returns:
            The number of pages for this version.
        """
        return row.pages.count()

    def render_num_images_column(self, row: Version, column: str) -> str:
        """
        Render our ``num_images`` column.  This is the number of
        :py:class:`sphinx_hosting.models.SphinxImage` objects imported for this
        version.

        Args:
            row: the ``Version`` we are rendering
            colunn: the name of the column to render

        Returns:
            The number of images for this version.
        """
        return row.images.count()



from typing import List

from django.urls import path, re_path, URLPattern

from .views import (
    ClassifierViewSet,
    GlobalSphinxPageSearchView,
    ProjectCreateView,
    ProjectDeleteView,
    ProjectDetailView,
    ProjectListView,
    ProjectUpdateView,
    SphinxPageDetailView,
    VersionDeleteView,
    VersionDetailView,
    VersionUploadView
)
from .wildewidgets import ProjectClassifierSelectorWidget


app_name: str = "sphinx_hosting"

urlpatterns: List[URLPattern] = [
    path('', ProjectListView.as_view(), name='project--list'),
    path('project/', ProjectCreateView.as_view(), name='project--create'),
    path('project/<str:slug>/', ProjectDetailView.as_view(), name='project--detail'),
    path('project/<str:slug>/update/', ProjectUpdateView.as_view(), name='project--update'),
    path('project/<str:slug>/upload/', VersionUploadView.as_view(), name='version--upload'),
    path('project/<str:slug>/delete/', ProjectDeleteView.as_view(), name='project--delete'),
    path('project/<str:project_slug>/<str:version>/', VersionDetailView.as_view(), name='version--detail'),
    path('project/<str:project_slug>/<str:version>/delete/', VersionDeleteView.as_view(), name='version--delete'),
    re_path(
        r'project/(?P<project_slug>[-a-zA-Z0-9_.]+)/(?P<version>[^/]+)/(?P<path>.*)/',
        SphinxPageDetailView.as_view(),
        name='sphinxpage--detail'
    ),
    path('search/', GlobalSphinxPageSearchView.as_view(), name='search'),
]
urlpatterns += ClassifierViewSet(url_prefix='lookups', url_namespace=app_name).get_urlpatterns()
urlpatterns += ProjectClassifierSelectorWidget.get_urlpatterns(url_namespace=app_name)



# Generated by Django 4.1.2 on 2023-02-04 01:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sphinxhostingcore', '0008_Version_archived'),
    ]

    operations = [
        migrations.AddField(
            model_name='sphinxpage',
            name='searchable',
            field=models.BooleanField(default=False, help_text='Should this page be included in the search index?', verbose_name='Searchable'),
        ),
        migrations.AlterField(
            model_name='sphinxpage',
            name='orig_body',
            field=models.TextField(blank=True, help_text='The original body for the page, extracted from the page JSON. Some pages have no body.We save this here in case we need to reprocess the body at some later date.', verbose_name='Body (Original)'),
        ),
    ]



from typing import Dict

from django.utils.text import slugify

from wildewidgets import (
    Block,
    CardWidget,
    CollapseWidget,
    CheckboxInputBlock,
    HorizontalLayoutBlock,
    LinkButton,
    NavLinkToggle,
    UnorderedList,
)

from ..models import Classifier, ClassifierNode


class ClassifierFilterForm(Block):
    """
    This is the tree-like classifier filter form that appears to the right of
    the :py:class:`sphinx_hosting.wildewidgets.project.ProjectTable`.  It
    is embedded in :py:class:`ClassifierFilterBlock`.

    It allows the user to select a set of classifiers by which to filter the
    projects listing table.
    """

    SCRIPT = """
$(document).ready(function() {{
    function classifier_filter_check_parents($li, state) {{
        var $siblings = $li.siblings();
        var $parent = $li.parent().closest('li');
        state = state && $siblings.children('label').find('input').prop('checked');
        $parent.children('label').find('input').prop('checked', state);
        if ($parent.parents('li').length) {{
            classifier_filter_check_parents($parent, state);
        }};
    }}
    function classifier_filter_open_sections() {{
        $('{block_id} input[type="checkbox"]:checked').each(function() {{
            var last = null;
            var target = null;
            var parents = $(this).parents().each(function() {{
                if ($(this).hasClass('classifiers__filter__form')) {{
                    target = last.prev();
                    return false;
                }}
                last = $(this);
            }});
            target.attr('aria-expanded', 'true');
            var collapse_id = target.attr('data-bs-target');
            $(collapse_id).addClass('show');
        }});
    }}
    classifier_filter_open_sections();
    $('{block_id} input').change(function () {{
        var $cb = $(this);
        var $li = $cb.closest('li');
        var state = $cb.prop('checked');

        // check all children
        $li.find('input').prop('checked', state);

        // check all parents, as applicable
        if ($li.parents('li').length) {{
            classifier_filter_check_parents($li, state);
        }};
    }});
    $('{block_id} .classifier--submit').on('click', function () {{
        var classifier_ids = [];
        $('{block_id} input[type="checkbox"]:checked').each(function() {{
            classifier_ids.push($(this).val());
        }});
        var table = $('.{table_name}').DataTable();
        table.column({column_number}).search(classifier_ids.join()).draw();
    }});
    $('{block_id} .classifier--clear').on('click', function () {{
        $('{block_id} input[type="checkbox"]:checked').each(function() {{
            $(this).prop('checked', false)
        }});
    }});
}});
"""

    block: str = 'classifiers__filter__form'
    tag: str = 'form'

    def __init__(self, table_name: str, column_number: int, **kwargs):
        self.table_name = table_name
        self.column_number = column_number
        super().__init__(**kwargs)
        self._css_id = self.block
        self.script = self.SCRIPT.format(
            block_id=f'#{self._css_id}',
            column_number=self.column_number,
            table_name=self.table_name
        )
        self._attributes['method'] = 'get'
        self._attributes['name'] = self.block
        self.tree: Dict[str, ClassifierNode] = Classifier.objects.tree()
        for node in self.tree.values():
            name = slugify(node.title)
            target = f'#{name}'
            self.add_block(NavLinkToggle(node.title, collapse_id=target, css_class='my-2'))
            contents = UnorderedList()
            self.add_block(CollapseWidget(contents, css_id=name))
            self.add_subtree(contents, node.items)
        self.add_block(
            HorizontalLayoutBlock(
                LinkButton(name='classifier--submit', text='Filter Projects', color='primary'),
                LinkButton(
                    name='classifier--clear',
                    text='Clear',
                    color='outline-secondary',
                    css_class="mt-2 mt-xl-0"
                ),
                css_class='my-3',
                justify='between',
                flex_size='xl'
            )
        )

    def add_subtree(self, contents: UnorderedList, nodes: Dict[str, ClassifierNode]) -> None:
        """
        Add a subtree of classifier checkboxes.

        Args:
            contents: the ``<ul>`` block to which to add our list of classifier checkboxes
            nodes (_type_): _description_
        """
        for node in nodes.values():
            checkbox = self.get_checkbox(node)
            if node.items:
                container = Block(tag='li')
                container.add_block(checkbox)
                sub_contents = UnorderedList()
                container.add_block(sub_contents)
                contents.add_block(container)
                self.add_subtree(sub_contents, node.items)
            else:
                contents.add_block(checkbox)

    def get_checkbox(self, node: ClassifierNode) -> HorizontalLayoutBlock:
        """
        Build and return the :py:class:`wildewidgets.CheckboxInputBlock` for the
        classifier ``node``.

        Args:
            node: the classifier data

        Returns:
            A configured :py:class:`wildewidgets.CheckboxInputBlock`
        """
        value = node.classifier.id if node.classifier else 'empty'
        return CheckboxInputBlock(
            label_text=node.title,
            bold=False,
            input_name=slugify(node.title),
            value=value
        )


class ClassifierFilterBlock(CardWidget):
    """
    A :py:class:`wildewidgets.CardWidget` that contains the
    :py:class:`ClassifierFilterForm`.  This the right of the
    :py:class:`sphinx_hosting.wildewidgets.project.ProjectTable`.  This widget
    is embedded in
    :py:class:`sphinx_hosting.wildewdigets.project.ProjectTableWidget`

    Args:
        table_name: the name of the dataTables table to control
        column_number: the number of the column in the dataTable that contains
            classifier names
    """

    name: str = 'classifiers__filter'

    card_title: str = 'Filter by classifier'

    def __init__(self, table_name: str, column_number: int, **kwargs):
        self.table_name = table_name
        self.column_number = column_number
        super().__init__(**kwargs)
        self.set_widget(
            ClassifierFilterForm(
                table_name=self.table_name,
                column_number=column_number
            )
        )
